package Git::Raw::Cert::X509;
$Git::Raw::Cert::X509::VERSION = '0.56'; # TRIAL
use strict;
use warnings;

=head1 NAME

Git::Raw::Cert::X509 - Git X509 certificate class

=head1 VERSION

version 0.56

=head1 DESCRIPTION

A L<Git::Raw::Cert::X509> object represents a X509 certificate.

B<WARNING>: The API of this module is unstable and may change without warning
(any change will be appropriately documented in the changelog).

=head1 METHODS

=head2 data( )

Data for the certificate.

=head1 AUTHOR

Jacques Germishuys <jacquesg@striata.com>

=head1 LICENSE AND COPYRIGHT

Copyright 2014 Jacques Germishuys.

This program is free software; you can redistribute it and/or modify it
under the terms of either: the GNU General Public License as published
by the Free Software Foundation; or the Artistic License.

See http://dev.perl.org/licenses/ for more information.

=cut

1; # End of Git::Raw::Cert::X509
