/*
 * Copyright (C) the libgit2 contributors. All rights reserved.
 *
 * This file is part of libgit2, distributed under the GNU GPL v2 with
 * a Linking Exception. For full terms see the included COPYING file.
 */
#ifndef INCLUDE_clone_h__
#define INCLUDE_clone_h__

extern int git_clone__should_clone_local(const char *url, git_clone_local_t local);

#endif
